"use strict";

var _interopRequireDefault = require("reshow-runtime/helpers/interopRequireDefault");
var _interopRequireWildcard = require("reshow-runtime/helpers/interopRequireWildcard");
exports.__esModule = true;
exports.DeferredActionUtil = void 0;
var _objectDestructuringEmpty2 = _interopRequireDefault(require("reshow-runtime/helpers/objectDestructuringEmpty"));
var _extends2 = _interopRequireDefault(require("reshow-runtime/helpers/extends"));
var _objectSpread2 = _interopRequireDefault(require("reshow-runtime/helpers/objectSpread2"));
var _objectWithoutPropertiesLoose2 = _interopRequireDefault(require("reshow-runtime/helpers/objectWithoutPropertiesLoose"));
var _getObjectValue = _interopRequireWildcard(require("get-object-value"));
var _setObjectValue = _interopRequireDefault(require("set-object-value"));
var _seturl = require("seturl");
var _callFunc = _interopRequireDefault(require("call-func"));
var _reshowConstant = require("reshow-constant");
var _getParams5 = require("./getParams.js");
var _getTime = _interopRequireDefault(require("./getTime.js"));
var _excluded = ["stop", "wait", "deferredAction", "deferredKey"]; // @ts-check
var deferredActionStorageKey = "deferredAction";
var PARAMS = "params";
var hashKey = "__hash";
var seqKey = "__seq";

/**
 * @template StateType
 * @template ActionType
 * @typedef {import("reshow-flux-base").DispatchFunction<StateType, ActionType>} DispatchFunction
 */
/**
 * @typedef {Object<sinon,any>} StateType
 */
/**
 * @typedef {import("reshow-flux-base").ActionObject} ActionObject
 */

/**
 * @template StateType
 * @template ActionType
 * @param {any} deferredActionMapObject
 * @param {DispatchFunction<StateType, ActionType>} dispatch
 */
var processDeferredAction = function processDeferredAction(deferredActionMapObject, dispatch) {
  /**
   * @param {any} deferredMapOrSeq
   * @param {number|string} key
   */
  var processDeferred = function processDeferred(deferredMapOrSeq, key) {
    var oneAction = deferredMapOrSeq[key];
    var _getParams = (0, _getParams5.getParams)(oneAction),
      wait = _getParams.wait,
      stop = _getParams.stop;
    if (!wait || wait <= 0) {
      if (!stop) {
        if (null == wait) {
          delete oneAction.params.wait;
        }
        dispatch && dispatch(oneAction);
      }
      delete deferredMapOrSeq[key];
    } else {
      oneAction.params.wait = wait - 1;
    }
    return deferredMapOrSeq[key];
  };
  var seq = (0, _getObjectValue["default"])(deferredActionMapObject, [seqKey]);
  if ((0, _reshowConstant.IS_ARRAY)(seq)) {
    deferredActionMapObject.__seq = seq.filter(function (_action, key) {
      return processDeferred(seq, key);
    });
  }
  var hash = (0, _getObjectValue["default"])(deferredActionMapObject, [hashKey]);
  if (hash) {
    (0, _reshowConstant.KEYS)(hash).forEach(function (key) {
      return processDeferred(hash, key);
    });
  }
  return deferredActionMapObject;
};

/**
 * @typedef {object} StorageType
 * @property {function(string):any} get
 * @property {function(string, any):any} set
 */

/**
 * @param {StorageType} storage
 */
var DeferredActionUtil = exports.DeferredActionUtil = function DeferredActionUtil(storage) {
  var getAllDeferredAction = function getAllDeferredAction() {
    return (0, _getObjectValue.toMap)(storage.get(deferredActionStorageKey));
  };
  var getOneDeferredAction = function getOneDeferredAction( /**@type string*/k) {
    return (0, _getObjectValue.toMap)(getAllDeferredAction().__hash)[k];
  };
  var updateDeferredAction = function updateDeferredAction( /**@type any*/deferredActionMapObject) {
    return storage.set(deferredActionStorageKey, deferredActionMapObject);
  };
  var removeDeferredAction = function removeDeferredAction( /**@type string*/key) {
    var deferredActionMapObject = getAllDeferredAction();
    if ((0, _getObjectValue["default"])(deferredActionMapObject, [hashKey, key])) {
      delete deferredActionMapObject.__hash[key];
      updateDeferredAction(deferredActionMapObject);
    }
  };

  /**
   * @param {any} action
   * @param {string} key
   */
  var getMergeWithDeferredAction = function getMergeWithDeferredAction(action, key) {
    var deferredActionMapObject = getAllDeferredAction();
    var _get = (0, _getObjectValue["default"])(deferredActionMapObject, [hashKey, key, PARAMS], {}),
      stop = _get.stop,
      wait = _get.wait,
      deferredAction = _get.deferredAction,
      deferredKey = _get.deferredKey,
      restParams = (0, _objectWithoutPropertiesLoose2["default"])(_get, _excluded);
    (0, _reshowConstant.KEYS)(restParams).forEach(function (pKey) {
      var p = restParams[pKey];
      var newP = _reshowConstant.OBJECT === typeof p ? (0, _objectSpread2["default"])((0, _objectSpread2["default"])({}, p), (0, _getObjectValue["default"])(action, [PARAMS, pKey], {})) : (0, _getObjectValue["default"])(action, [PARAMS, pKey], p);
      (0, _setObjectValue["default"])(action, [PARAMS, pKey], newP);
    });
    delete action.params.mergeWithDeferredKey;
    return action;
  };

  /**
   * @param {any} action
   * @param {string=} key
   */
  var pushDeferredAction = function pushDeferredAction(action, key) {
    var _getParams2 = (0, _getParams5.getParams)(action),
      params = (0, _extends2["default"])({}, ((0, _objectDestructuringEmpty2["default"])(_getParams2), _getParams2));
    var thisAction = {
      params: params,
      type: action.type
    };
    (0, _setObjectValue["default"])(thisAction, [PARAMS, "deferredAction"], {
      from: (0, _seturl.url)(),
      time: (0, _getTime["default"])().toString()
    });
    var deferredActionMapObject = getAllDeferredAction();
    if (key) {
      (0, _setObjectValue["default"])(deferredActionMapObject, [hashKey, key], thisAction);
    } else {
      (0, _setObjectValue["default"])(deferredActionMapObject, [seqKey], thisAction, true);
    }
    updateDeferredAction(deferredActionMapObject);
  };

  /**
   * @template StateType
   * @template ActionType
   * @param {DispatchFunction<StateType, ActionType>} dispatch
   */
  var process = function process(dispatch) {
    return updateDeferredAction(processDeferredAction(getAllDeferredAction(), dispatch));
  };

  /**
   * @param {StateType} state
   * @param {ActionObject} action
   */
  var handleAction = function handleAction(state, action) {
    var _getParams3 = (0, _getParams5.getParams)(action),
      mergeWithDeferredKey = _getParams3.mergeWithDeferredKey;
    if (mergeWithDeferredKey) {
      action = getMergeWithDeferredAction(action, mergeWithDeferredKey);
    }
    var actionHandler = state.get("deferredActionHandler");
    var next = (0, _callFunc["default"])(actionHandler, [state, action]) || state;
    var _getParams4 = (0, _getParams5.getParams)(action),
      wait = _getParams4.wait,
      stop = _getParams4.stop,
      deferredKey = _getParams4.deferredKey; // need locate after next
    if (_reshowConstant.T_NULL == wait && !stop) {
      if (mergeWithDeferredKey && mergeWithDeferredKey !== deferredKey) {
        removeDeferredAction(mergeWithDeferredKey);
      }
    }
    return next;
  };
  return {
    process: process,
    handleAction: handleAction,
    getAll: getAllDeferredAction,
    getOne: getOneDeferredAction,
    push: pushDeferredAction
  };
};