import _objectDestructuringEmpty from "reshow-runtime/es/helpers/objectDestructuringEmpty";
import _extends from "reshow-runtime/es/helpers/extends";
import _objectSpread from "reshow-runtime/es/helpers/objectSpread2";
import _objectWithoutProperties from "reshow-runtime/es/helpers/objectWithoutProperties";
var _excluded = ["stop", "wait", "deferredAction", "deferredKey"];
// @ts-check
import get, { toMap } from "get-object-value";
import set from "set-object-value";
import { url } from "seturl";
import callfunc from "call-func";
import { T_NULL, OBJECT, KEYS, IS_ARRAY } from "reshow-constant";
import { getParams } from "./getParams.mjs";
import getTime from "./getTime.mjs";
var deferredActionStorageKey = "deferredAction";
var PARAMS = "params";
var hashKey = "__hash";
var seqKey = "__seq";

/**
 * @template StateType
 * @template ActionType
 * @typedef {import("reshow-flux-base").DispatchFunction<StateType, ActionType>} DispatchFunction
 */
/**
 * @typedef {Object<sinon,any>} StateType
 */
/**
 * @typedef {import("reshow-flux-base").ActionObject} ActionObject
 */

/**
 * @template StateType
 * @template ActionType
 * @param {any} deferredActionMapObject
 * @param {DispatchFunction<StateType, ActionType>} dispatch
 */
var processDeferredAction = function processDeferredAction(deferredActionMapObject, dispatch) {
  /**
   * @param {any} deferredMapOrSeq
   * @param {number|string} key
   */
  var processDeferred = function processDeferred(deferredMapOrSeq, key) {
    var oneAction = deferredMapOrSeq[key];
    var {
      wait,
      stop
    } = getParams(oneAction);
    if (!wait || wait <= 0) {
      if (!stop) {
        if (null == wait) {
          delete oneAction.params.wait;
        }
        dispatch && dispatch(oneAction);
      }
      delete deferredMapOrSeq[key];
    } else {
      oneAction.params.wait = wait - 1;
    }
    return deferredMapOrSeq[key];
  };
  var seq = get(deferredActionMapObject, [seqKey]);
  if (IS_ARRAY(seq)) {
    deferredActionMapObject.__seq = seq.filter(function (_action, key) {
      return processDeferred(seq, key);
    });
  }
  var hash = get(deferredActionMapObject, [hashKey]);
  if (hash) {
    KEYS(hash).forEach(function (key) {
      return processDeferred(hash, key);
    });
  }
  return deferredActionMapObject;
};

/**
 * @typedef {object} StorageType
 * @property {function(string):any} get
 * @property {function(string, any):any} set
 */

/**
 * @param {StorageType} storage
 */
export var DeferredActionUtil = function DeferredActionUtil(storage) {
  var getAllDeferredAction = function getAllDeferredAction() {
    return toMap(storage.get(deferredActionStorageKey));
  };
  var getOneDeferredAction = function getOneDeferredAction( /**@type string*/k) {
    return toMap(getAllDeferredAction().__hash)[k];
  };
  var updateDeferredAction = function updateDeferredAction( /**@type any*/deferredActionMapObject) {
    return storage.set(deferredActionStorageKey, deferredActionMapObject);
  };
  var removeDeferredAction = function removeDeferredAction( /**@type string*/key) {
    var deferredActionMapObject = getAllDeferredAction();
    if (get(deferredActionMapObject, [hashKey, key])) {
      delete deferredActionMapObject.__hash[key];
      updateDeferredAction(deferredActionMapObject);
    }
  };

  /**
   * @param {any} action
   * @param {string} key
   */
  var getMergeWithDeferredAction = function getMergeWithDeferredAction(action, key) {
    var deferredActionMapObject = getAllDeferredAction();
    var _get = get(deferredActionMapObject, [hashKey, key, PARAMS], {}),
      {
        stop,
        wait,
        deferredAction,
        deferredKey
      } = _get,
      restParams = _objectWithoutProperties(_get, _excluded);
    KEYS(restParams).forEach(function (pKey) {
      var p = restParams[pKey];
      var newP = OBJECT === typeof p ? _objectSpread(_objectSpread({}, p), get(action, [PARAMS, pKey], {})) : get(action, [PARAMS, pKey], p);
      set(action, [PARAMS, pKey], newP);
    });
    delete action.params.mergeWithDeferredKey;
    return action;
  };

  /**
   * @param {any} action
   * @param {string=} key
   */
  var pushDeferredAction = function pushDeferredAction(action, key) {
    var _getParams = getParams(action),
      params = _extends({}, (_objectDestructuringEmpty(_getParams), _getParams));
    var thisAction = {
      params,
      type: action.type
    };
    set(thisAction, [PARAMS, "deferredAction"], {
      from: url(),
      time: getTime().toString()
    });
    var deferredActionMapObject = getAllDeferredAction();
    if (key) {
      set(deferredActionMapObject, [hashKey, key], thisAction);
    } else {
      set(deferredActionMapObject, [seqKey], thisAction, true);
    }
    updateDeferredAction(deferredActionMapObject);
  };

  /**
   * @template StateType
   * @template ActionType
   * @param {DispatchFunction<StateType, ActionType>} dispatch
   */
  var process = function process(dispatch) {
    return updateDeferredAction(processDeferredAction(getAllDeferredAction(), dispatch));
  };

  /**
   * @param {StateType} state
   * @param {ActionObject} action
   */
  var handleAction = function handleAction(state, action) {
    var {
      mergeWithDeferredKey
    } = getParams(action);
    if (mergeWithDeferredKey) {
      action = getMergeWithDeferredAction(action, mergeWithDeferredKey);
    }
    var actionHandler = state.get("deferredActionHandler");
    var next = callfunc(actionHandler, [state, action]) || state;
    var {
      wait,
      stop,
      deferredKey
    } = getParams(action); // need locate after next
    if (T_NULL == wait && !stop) {
      if (mergeWithDeferredKey && mergeWithDeferredKey !== deferredKey) {
        removeDeferredAction(mergeWithDeferredKey);
      }
    }
    return next;
  };
  return {
    process,
    handleAction,
    getAll: getAllDeferredAction,
    getOne: getOneDeferredAction,
    push: pushDeferredAction
  };
};